/*
 * @(#)%Name% %Version% %Time%
 * 
 * Copyright (c) 1997 Rutgers University
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that (i) the above copyright notice and this permission notice
 * appear in all copies of the software and related documentation. (ii) 
 * the name of Rutgers University may not be used in any advertising
 * or publicity relating to the software without the specific, prior
 * written permission of Rutgers University.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL RUTGERS UNIVERSITY BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 * CopyrightVersion %Version%
 * @author		%Author%
 * @version		%Version%, %Time%
 * @since		ACT %Version%
 *
 */

// last modified 01/20/98

package EDU.rutgers.CORBA.GIOP.messages;

import ORG.omg.CORBA.OctetSequence;
import ORG.omg.CORBA.Principal;
import ORG.omg.CORBA.EMPTYByteArray;

import EDU.rutgers.act.presentation.TypeReader;
import EDU.rutgers.act.presentation.TypeWriter;

import EDU.rutgers.CORBA.IOP.ServiceContext;

/**
 * [CORBA 2.0 Specification 12-17]
 *
 * @see GIOPMessage
 *
 * @see GIOPReplyMessage
 * @see GIOPCancelRequestMessage
 * @see GIOPLocateRequestMessage
 * @see GIOPLocateReplyMessage
 * @see GIOPCloseConnectionMessage
 * @see GIOPMessageErrorMessage
 *
 * @since ACT 0.1
 */
public final class GIOPRequestMessage extends GIOPMessage
{
    /**
     * Contains the ORB service context passed from the client to the server
     *
     * @see EDU.rutgers.act.IOP.ServiceContext
     *
     * @since ACT 0.1
     */
	public ServiceContext[] service_context_;

    /**
     * true if the request is expected to have an associated reply
     * false if the operation is defined as oneway or if the operation is invoked with DII
     * and the invocation flags include the INV_NO_RESPONSE flag.
     *
     * @since ACT 0.1
     */
	public boolean response_expected_;

    /**
     * Identifies the object which is the target of the invocation. It is from the IOR 
     * encapsulated IIOP ProfileBody.
     * It is sequence <octet>, but aslo could be an encapsulation
     * by now, we choose to use sequence <octet>.
     *
     * @see EDU.rutgers.act.IOP.IOR
     * @see EDU.rutgers.act.IIOP.ProfileBody
     *
     * @since ACT 0.1
     */
	public byte[] object_key_;

    /**
     * Contains the name of the operation being invoked
     *
     * @since ACT 0.1
     */
	public String operation_;

    /**
     * Contans a value identifying the request principal
     *
     * @see ORG.omg.CORBA.Principal
     *
     * @since ACT 0.1
     */
	public Principal requesting_principal_;

    private byte[] request_body_ = EMPTYByteArray.EMPTY;

    private int requestHeaderSize_; // network size
        
    /**
     * Empty constructor, it could only be used by the GIOPMessageFactory
     *
     * @since ACT 0.1
     */
    protected GIOPRequestMessage() {
        super(MessageType.Request);
    }

    /**
     * @param ServiceContext[] service_context
	 * @param int request_id
	 * @param boolean response_expected
	 * @param byte[] object_key
	 * @param String operation 
	 * @param Principal requesting_principal
     *
     * @since ACT 0.1
     */
    public GIOPRequestMessage(
		ServiceContext[] service_context, 
		int request_id, 
		boolean response_expected, 
		byte[] object_key, 
		String operation, 
		Principal requesting_principal) {
        super(MessageType.Request);

        service_context_ = service_context;
        request_id_ = request_id;
        response_expected_ = response_expected;
        object_key_ = object_key;
        operation_ = operation;
        requesting_principal_ = requesting_principal;

        calRequestHeaderSize();
    }

    /**
     * Calculate the header network size. In case the user just resuse this object
     * by replacing some member fields, this method forces to reset the GIOP header
     * message_size field.
     *
     * @since ACT 0.1
     */
    public void calRequestHeaderSize() {
        int twoHeadersSize = super.giopHeaderSize_;

        // service_context_[].length as unsigned long
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 4) + 4; 
		int len = service_context_.length;
		for (int i = len; i-- != 0;) {
            // service_context_.context_id and service_context_.length as unsigned long
			twoHeadersSize += (super.ts_.calAlignGap(twoHeadersSize, 4) + 4) << 1;
            // service_context_.context_data_ as byte array
            len = service_context_[len - i - 1].context_data_.length;
            twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 1) + len; 
		}

        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 4) + 4; // request_id
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 1) + 1; // response_expected

        // object_key_.length as unsigned long
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 4) + 4;
        len = object_key_.length;
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 1) + len; // object_key_
        
        // operation.length as unsigned long
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 4) + 4;
        len = operation_.length();
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 1) + len // operation_ // CRIS!!!
	+ 1;
        
        // requesting_principal.length as unsigned long
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 4) + 4;
        len = requesting_principal_.values().length;
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 1) + len; // requesting_principal

        requestHeaderSize_ = twoHeadersSize - super.giopHeaderSize_;

        super.setMessageSize(requestHeaderSize_ + request_body_.length);
    }

    /**
     * Must be called before setting any Request body.
     * The user should use this value to determine the Request body alignment.
     * This alignment is based on the GIOPMessage.ts_ transfer syntax.
     *
     * @return int size of the GIOP MessageHeader and LocateReplyHeader 
     * after the corresponding transfer syntax
     *
     * @see GIOPMessage#ts_
     * @see #setRequestBody
     * 
     * @since ACT 0.1
     */
    public int getBeforeBodySize() {
        return requestHeaderSize_ + super.giopHeaderSize_;
    }

    /**
     * Set the Request body.
     * Normally, the user should first call getBeforeBodySize to get all headers size,
     * then format the body with that size as the index.
     *
     * Here is the sample code to compose a RequestMessage.
     *
     * ...
     * // create the locate reply message
     * GIOPRequestMessage requestMsg = new GIOPRequestMessage(...);
     *
     * // create the steam and writer
     * ByteArrayOutputStream bodyStream = new ByteArrayOutputStream();
     * TypeWriter bodyWriter = TransferSyntaxFactory.createTypeWriter("CDR", new
     *                               OutputStreamWriter(bodyStream));
     *               
     * // reset the writer
     * bodyWriter.setByteOrder(replyMsg.byte_order_);
     * bodyWriter.setIndex(replyMsg.getBeforeBodySize());
     *
     * // write the necessary fields into the TypeWriter
     * ...
     * bodyWriter.flush(); // finished
     * byte[] requestBody = bodyStream.toByteArray();
     *                
     * // set the reply body
     * requestMsg.setRequestBody(requestBody);
     * ...
     *
     * @param byte[] request body
     *
     * @see #getBeforeBodySize
     * @see #getRequestBody
     * 
     * @since ACT 0.1
     */
    public void setRequestBody(byte[] request_body) {
        if (request_body == null) {
            return;
        }
        request_body_ = request_body;
        super.setMessageSize(requestHeaderSize_ + request_body_.length);
    }

    /**
     * Retrieve the request body. This body is only byte array and it is 
     * aligned at the beginning of the GIOP MessageHeader.
     *
     * @return byte[] body
     *
     * @see #setRequestBody
     *
     * @since ACT 0.1
     */
    public byte[] getRequestBody() {
        return request_body_;
    }

    /**
     * Read from the TypeReader stream
     *
     * @exception java.io.IOException If any IO error occurs
     *
     * @see ORG.omg.CORBA.Streamable#read
     * @see EDU.rutgers.act.presentation.TypeReader
     *
     * @since ACT 0.1
     */
    public void read(TypeReader in)
        throws java.io.IOException {
        super.read(in);

		int len = in.readULong();
		service_context_ = new ServiceContext[len];
		for (int i = len; i-- != 0; ) {
            service_context_[len - i - 1] = new ServiceContext();
			service_context_[len - i - 1].read(in);
		}	

        request_id_ = in.readULong();

        response_expected_ = in.readBoolean();

		len = in.readULong();
		object_key_ = new byte[len];
		in.readOctetArray(object_key_, 0, len);

        operation_ = in.readString();

        requesting_principal_ = new Principal();
        requesting_principal_.read(in);

        // request body
        int twoHeadersSize = in.getIndex();   
        requestHeaderSize_ = twoHeadersSize - super.giopHeaderSize_;
        int bodySize = super.message_size_ - requestHeaderSize_;
        request_body_ = new byte[bodySize];
        in.readOctetArray(request_body_, 0, bodySize);
    }

    /**
     * Write to the TypeWriter stream.
     *
     * @exception java.io.IOException If any IO error occurs
     *
     * @see ORG.omg.CORBA.Streamable#write
     * @see EDU.rutgers.act.presentation.TypeWriter
     *
     * @since ACT 0.1
     */
    public void write(TypeWriter out)
        throws java.io.IOException {
        super.write(out);

		int len = service_context_.length;
        out.writeULong(len);
		for(int i = len; i-- != 0;) {
			service_context_[len - i - 1].write(out);
		}	

        out.writeULong(request_id_);
        out.writeBoolean(response_expected_);

        len = object_key_.length;
        out.writeULong(len);
		out.writeOctetArray(object_key_, 0, len);

        out.writeString(operation_);
        requesting_principal_.write(out);

        // request body
        out.writeOctetArray(request_body_, 0, request_body_.length);
	}

// Object
/*
    public boolean equals(Object o) {
		if(super.equals(o) && (o != null) && (o instanceof GIOPRequestMessage)) {
            if (o == this) {
                return true;
            }
            GIOPRequestMessage msg = (GIOPRequestMessage)o;
            int len = service_context_.length;
            if (len != msg.service_context_.length) {
                return false;
            }
		    for (int i = len; i-- != 0; ) {
                if (service_context_[len - i].equals(msg.service_context_[len - i]) == false) {
                    return false;
                }
		    }
            if (request_id_ == msg.request_id_ &&
                response_expected_ == msg.response_expected_ &&
                String.valueOf(object_key_).equals(String.valueOf(msg.object_key_)) &&
                operation_.equals(msg.operation_) &&
                requesting_principal_.equals(msg.requesting_principal_)) {
                return true;
            }
        }
        return false;
    }
*/

    /**
     * Returns a string representation of the GIOPRequestMessage. 
     *
     * @return String GIOPRequestMessage description
     *
     * @since ACT 0.1
     */
    public String toString() {
        StringBuffer buf = new StringBuffer(super.toString()).
            append("\nGIOP:GIOPRequestMessage").
            append("\n\t[Header] ");

        int len = service_context_.length;
		for (int i = len; i-- != 0; ) {
			buf.append("\n\t").append(service_context_[len - i - 1].toString());
		}
		buf.append("\n\trequest_id: ").append(request_id_).
            append("\n\tresponse_expected: ").append(response_expected_).
            append("\n\tobject_key: ").append(OctetSequence.byteArray2String(object_key_)).
            append("\n\toperation: ").append(operation_).
            append("\n\trequesting_principal: ").append(requesting_principal_.toString()).
            append("\n\t[Body]\n\t").append(OctetSequence.byteArray2String(request_body_));
		return String.valueOf(buf);
      }
}
