package EDU.rutgers.cbus.protocol.messages;

import java.util.EventObject;
import java.io.*;
import java.util.StringTokenizer;
import EDU.rutgers.act.endpoint.Endpoint;
import EDU.rutgers.act.protocol.Message;
import EDU.rutgers.act.presentation.*;

public class CBusEvent extends EventObject implements Message, Cloneable
{
  protected int objKey_;        // reference for identifying the destination
  protected EventObject event_; // keeps the actual event to be sent
  protected Endpoint endpoint_; // keeps the destination of this message  

  protected int userID_;
  protected int timestamp_;
  protected int seqNum_;
  protected short priority1_;
  protected short priority2_;
  protected boolean useCC_;
  protected boolean undo_;

  protected byte[] body_;

  // used by CBusEvent creators only (normally the client/server layer)
  public CBusEvent(int userID, EventObject event)
    {
      super(event.getSource());
      userID_ = userID;
      event_ = event;
      serialize();
    }
	
  public CBusEvent(EventObject event)
    {
      super(event.getSource());
      event_ = event;
      serialize();
    }

  CBusEvent(Object source) // used by the factory only (CBusEventFactory)
    {
      super(source);
    }
  
  protected void serialize()
    {
      // get the serialized body of this message
      body_ = null;
      try
	{
	  ByteArrayOutputStream os = new ByteArrayOutputStream();
	  ObjectOutput oo = new ObjectOutputStream(os);
	  oo.writeObject(event_);
	  oo.close();
	  body_ = os.toByteArray();
	}
      catch(java.io.IOException e)
	{
	  System.out.println(
	    "EDU.rutgers.cbus.protocol.CBusEvent: error serializing message body" + e);
	}
    }

	public int getUserID()
	{
		return userID_;
	}

	public void setTimestamp(int timestamp)
	{
	    timestamp_ = timestamp;    
	}
	
	public int getTimestamp()
	{
	    return timestamp_;
	}
	
	public void setSeqNum(int seqNum)
	{
	    seqNum_ = seqNum;
	}
	
	public int getSeqNum()
	{
	    return seqNum_;
	}
	
	public void setPriority(String priority)
	{
	    StringTokenizer st = new StringTokenizer(priority, ".");
	    priority1_ = new Short(st.nextToken()).shortValue();
	    priority2_ = new Short(st.nextToken()).shortValue();
	}
	
	public String getPriority()
	{
	    return String.valueOf(priority1_) + "." + String.valueOf(priority2_);
	}
	
	public short getPriority1()
	{
	    return priority1_;
	}
	
	public short getPriority2()
	{
	    return priority2_;
	}
	
  public int getObjKey()
    {
      return objKey_;
    }
  
  public void setObjKey(int objKey)
    {
      objKey_ = objKey;
    }
  
	public void setUseCC(boolean useCC)
	{
	    useCC_ = useCC;
	}
	    
	public boolean getUseCC()
	{
	    return useCC_;
	}
	
	public void setUndo(boolean undo)
	{
	    undo_ = undo;
	}
	
	public boolean getUndo()
	{
	    return undo_;
	}
	    
  public EventObject getUserEvent()
    {
      return event_;
    }

	public EventObject getEvent()
	{
		return event_;
	}
	
		public boolean hasLowerPriorityThan(CBusEvent cbe)
	{
		boolean result = false;

		if(priority1_ < cbe.getPriority1())
		{
			result = true;
		}
		else if(priority1_ == cbe.getPriority1() &&
			priority2_ < cbe.getPriority2())
		{
			result = true;
		}

		return result;
	}

	public boolean hasHigherPriorityThan(CBusEvent cbe)
	{
		boolean result = false;

		if(priority1_ > cbe.getPriority1())
		{
			result = true;
		}
		else if(priority1_ == cbe.getPriority1() &&
			priority2_ > cbe.getPriority2())
		{
			result = true;
		}

		return result;
	}

  public void setEndpoint(Endpoint ep)
    {
      endpoint_ = ep;
    }


  //
  // implementation of Message interface methods
  //
  public int getType()
    {
      // no types defined, yet
      return 0;
    }
  
  public int getTransactionId()
    {
      return 0;
    }
  
  public Endpoint getEndpoint()
    {
      return endpoint_;
    }
  
  public void read(TypeReader in) throws java.io.IOException
    {
      // start reading the message from the network
      
      objKey_ = (int)in.readUShort(); // read the object key

      userID_ = in.readULong();
      timestamp_ = in.readULong();
      seqNum_ = in.readULong();
      priority1_ = in.readShort();
      priority2_ = in.readShort();
      useCC_ = in.readBoolean();
      undo_ = in.readBoolean();

      // read the message body (the generated event)
      int len = in.readULong();
      body_ = new byte[len];
      in.readOctetArray(body_, 0, len);

      // extract the real message from the serialized body
      try
	{
	  ByteArrayInputStream is = new ByteArrayInputStream(body_);
	  ObjectInput oi = new ObjectInputStream(is);
	  event_ = (EventObject)oi.readObject();
	  oi.close();
	}
      catch(IOException e)
	{
	  System.out.println(
	    "EDU.rutgers.cbus.protocol.CBusEvent: error deserializing message body" + e);
	}
      catch(ClassNotFoundException e)
	{
	  System.out.println(
	    "EDU.rutgers.cbus.protocol.CBusEvent: error deserializing message body" + e);
	}
    }

  public void write(TypeWriter out) throws java.io.IOException
    {
      // start writing the message to the network
      
      out.writeUShort((short)objKey_); // write the object key
      out.writeULong(userID_);
      out.writeULong(timestamp_);
      out.writeULong(seqNum_);
      out.writeShort(priority1_);
      out.writeShort(priority2_);
      out.writeBoolean(useCC_);
      out.writeBoolean(undo_);

      // write the message body (the generated event)
      out.writeULong(body_.length);
      out.writeOctetArray(body_, 0, body_.length);
    }

  public synchronized Object clone()
    {
      try
	{
	  return super.clone();
	}
      catch(Exception e)
	{
	  System.out.println("@CBusEvent.clone(): " + e);
	  return null;
	}
    }
}
